#!/usr/bin/env python

import io
import os
import re
import math
from PIL import Image

from pycasso.shuffleseed import shuffle, unshuffle


class Canvas:
    def __init__(self, img, slice_size, seed=None):
        self.img = Image.open(img)
        self.slice_width, self.slice_height = slice_size
        self.seed = seed
        self.canvas = Image.new(
            mode="RGBA",
            size=(self.img_width, self.img_height),
            color=(255, 255, 255)
        )

        if not slice_size:
            raise ValueError(
                "Invalid slice size specified: input must be "
                "greater than or equal to one."
            )

    @property
    def img_width(self):
        return self.img.size[0]

    @property
    def img_height(self):
        return self.img.size[1]

    @property
    def img_filename(self):
        return self.get_basename(self.img.filename)

    @property
    def total_parts(self):
        return math.ceil(
            self.img_width / self.slice_width
        ) * math.ceil(
            self.img_height / self.slice_height
        )

    def get_slices(self):
        slices = {}
        vertical_slices = math.ceil(self.img_width / self.slice_width)

        for i in range(0, self.total_parts):
            slice = {}
            row = int(i / vertical_slices)
            col = i - row * vertical_slices
            slice['x'] = col * self.slice_width
            slice['y'] = row * self.slice_height
            slice['width'] = (
                self.slice_width - (
                    0 if (
                        slice['x'] + self.slice_width <= self.img_width
                    ) else (
                        slice['x'] + self.slice_width
                    ) - self.img_width
                )
            )
            slice['height'] = (
                self.slice_height - (
                    0 if (
                        slice['y'] + self.slice_height <= self.img_height
                    ) else (
                        slice['y'] + self.slice_height
                    ) - self.img_height
                )
            )

            if f"{slice['width']}-{slice['height']}" not in slices.keys():
                slices[f"{slice['width']}-{slice['height']}"] = []
            slices[f"{slice['width']}-{slice['height']}"].append(slice)

        return slices

    def get_cols_in_group(self, slices):
        if len(slices) == 1:
            return 1

        t = 'init'
        for i in range(0, len(slices)):
            if t == 'init':
                t = slices[i]['y']
            if t != slices[i]['y']:
                return i
                break

        return i if (self.img_height % self.slice_height) == 0 else i + 1

    def get_group(self, slices):
        this = {}
        this['slices'] = len(slices)
        this['cols'] = self.get_cols_in_group(slices)
        this['rows'] = len(slices) / this['cols']
        this['width'] = slices[0]['width'] * this['cols']
        this['height'] = slices[0]['height'] * this['rows']
        this['x'] = slices[0]['x']
        this['y'] = slices[0]['y']
        return this

    @staticmethod
    def get_basename(filename):
        name, ext = os.path.splitext(os.path.basename(filename))
        return name

    @staticmethod
    def set_stdname(filename):
        pattern = re.compile(r'[?"|:<>*]', flags=re.VERBOSE)
        return pattern.sub("", str(filename))

    def set_directory(self, path):
        if not os.path.isabs(path):
            path = os.path.join(os.getcwd(), self.set_stdname(path))

        os.makedirs(os.path.dirname(path), exist_ok=True)
        return path

    def export(self, mode=None, path=None, format=None):
        try:
            if not mode:
                mode = "scramble"

            if not format:
                format = "png"

            if format == "jpg":
                format = "jpeg"

            if format == "jpeg":
                self.canvas = self.canvas.convert("RGB")

            slices = self.get_slices()

            for g in slices:
                group = self.get_group(slices[g])
                shuffle_ind = []
                for i in range(0, len(slices[g])):
                    shuffle_ind.append(i)
                if mode == 'unscramble':
                    shuffle_ind = unshuffle(shuffle_ind, self.seed)
                if mode == 'scramble':
                    shuffle_ind = shuffle(shuffle_ind, self.seed)

                for i in range(0, len(slices[g])):
                    s = shuffle_ind[i]
                    row = int(s / group['cols'])
                    col = s - row * group['cols']
                    x = col * slices[g][i]['width']
                    y = row * slices[g][i]['height']

                    tiles = self.img.crop(
                        box=(
                            group['x'] + x,
                            group['y'] + y,
                            group['x'] + x + slices[g][i]['width'],
                            group['y'] + y + slices[g][i]['height']
                        )
                    )
                    self.canvas.paste(
                        tiles,
                        box=(
                            slices[g][i]['x'],
                            slices[g][i]['y'],
                            slices[g][i]['x'] + slices[g][i]['width'],
                            slices[g][i]['y'] + slices[g][i]['height']
                        )
                    )

            img_bytes = io.BytesIO()
            self.canvas.save(img_bytes, format)

            if path:
                if path.endswith('/'):
                    filename = f"pycasso-{self.img_filename}"
                    path = os.path.join(path, filename)

                output = f"{path}.{format}"

                with open(
                    self.set_directory(output)
                    if not os.path.exists(output) else output, 'wb'
                ) as file:
                    file.write(img_bytes.getvalue())

            return img_bytes

        except ValueError as error:
            raise SystemExit(f"Error: {error}")

    def close(self):
        self.img.close()



import math

from pycasso.prng import seedrandom


def islist(this):
    return type(this) is list


def seedrand(func, min, max):
    return math.floor(func() * (max - min + 1)) + min


def shuffle(list, seed=None):
    if not islist(list):
        return None

    size = len(list)
    rng = seedrandom(seed)
    resp = []
    keys = []

    for i in range(0, size):
        keys.append(i)

    for i in range(0, size):
        r = seedrand(rng, 0, len(keys)-1)
        g = keys[r]
        splice(keys, r, 1)
        resp.append(list[g])

    return resp


def unshuffle(list, seed=None):
    if not islist(list):
        return None

    size = len(list)
    rng = seedrandom(seed)
    resp = []
    keys = []

    for i in range(0, size):
        resp.append(None)
        keys.append(i)

    for i in range(0, size):
        r = seedrand(rng, 0, len(keys)-1)
        g = keys[r]
        splice(keys, r, 1)
        resp[g] = list[i]

    return resp


def splice(target, start, delete_count=None, *items):
    if delete_count is None:
        delete_count = len(target) - start

    total = start + delete_count
    removed = target[start:total]
    target[start:total] = items

    return removed



from time import time

from pycasso.constants import (
    width,
    chunks,
    startdenom,
    significance,
    overflow,
    mask
)
from pycasso.cipher import ARC4


def seedrandom(seed):
    key = []
    shortseed = mixkey(seed, key)
    arc4 = ARC4(shortseed)
    arc4.main()

    def prng():
        n = arc4.g(chunks)
        d = startdenom
        x = 0

        while n < significance:
            n = (n + x) * width
            d *= width
            x = arc4.g(1)

        while n >= overflow:
            n /= 2
            d /= 2
            x = x >> 1

        o = (n + x) / d

        return o

    return prng


def mixkey(seed, key):
    if not seed:
        seed = time()

    for j in range(0, len(str(seed))):
        key.insert(mask & j, ord(str(seed)[j]))

    return key



#!/usr/bin/env python

import argparse
import logging

from pycasso.unscramble import Canvas

log = logging.getLogger("pycasso.cli")


def main():
    try:
        parser = construct_parser()
        args = parser.parse_args()
        if not args.seed:
            log.warning(
                "No seed specified: this will automatically use "
                "the current system time."
            )
        if args.output.endswith('/'):
            log.warning(
                "No file name found: appending base file name to "
                "specified directory."
            )
        if args.n:
            width, height = args.n

        img = Canvas(
            args.input,
            slice_size=(width, height),
            seed=args.seed
        )
        img.export(
            mode=args.mode,
            path=args.output,
            format=args.format
        )

    except Exception as error:
        parser.error(error)


def construct_parser():
    parser = argparse.ArgumentParser(
        prog="pycasso",
        description="Split image into tiles and "
                    "scramble/unscramble them with seed.",
        epilog="Report bugs and make feature requests at "
               "https://github.com/catsital/pycasso/issues",
        add_help=False,
    )

    required = parser.add_argument_group("Required")
    required.add_argument("input", help="Path to image file.")
    required.add_argument("output", help="Output image file.")
    required.add_argument(
        "mode",
        choices=["scramble", "unscramble"],
        help="Modes available: scramble, unscramble"
    )

    optional = parser.add_argument_group("Optional")
    optional.add_argument(
        "-n",
        type=int,
        nargs=2,
        metavar=("WIDTH", "HEIGHT"),
        default=(5, 5),
        help="Size (width, height) of tiles to slice an image."
    )
    optional.add_argument(
        "-s",
        "--seed",
        type=str,
        default=None,
        help="Seed to use to scramble or unscramble an image."
    )
    optional.add_argument(
        "-f",
        "--format",
        type=str,
        default="png",
        help="Image format."
    )

    info = parser.add_argument_group("Info")
    info.add_argument(
        "-h",
        "--help",
        action="help",
        help="Display this screen."
    ),
    info.add_argument(
        "-v",
        "--version",
        action="version",
        help="Show program version.",
        version="%(prog)s 2.1.1"
    )

    return parser


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup_requirements = [
    "wheel>=0.35.1",
]

requirements = [
    "Pillow>=7.2.0",
    "Flask==2.0.2",
]

test_requirements = [
    "flake8>=3.8.3",
    "pytest>=5.4.3",
    "blinker==1.4",
]

dev_requirements = [
    *setup_requirements,
    *test_requirements,
]

extra_requirements = {
    "setup": setup_requirements,
    "test": test_requirements,
    "all": [*requirements, *dev_requirements,],
}

setup(
    name="image-scramble",
    version="2.1.1",
    author="catsital",
    author_email="catshital@gmail.com",
    description="Split image into tiles and scramble/unscramble them with seed.",
    entry_points={"console_scripts": ["pycasso=pycasso.__main__:main"],},
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/catsital/pycasso",
    project_urls={
        "Bug Tracker": "https://github.com/catsital/pycasso/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    extras_require=extra_requirements,
    zip_safe=False
)



from pycasso.constants import width, mask


class ARC4:
    def __init__(self, key):
        self.key = key

        global keylen, t, i, j, s
        keylen = len(self.key)
        t = None
        i = 0
        j = self.i = self.j = 0
        s = self.S = []

    def main(self):
        global i, j

        while i < width:
            s.append(i)
            i += 1

        for i in range(width):
            t = s[i]
            j = mask & (j + self.key[i % keylen] + t)
            s[i] = s[j]
            s[j] = t

        self.g(width)

    def g(self, count):
        t = None
        r = 0
        i = self.i
        j = self.j
        s = self.S

        while count:
            i = mask & (i + 1)
            t = s[i]
            j = mask & (j + t)
            s[i] = s[j]
            s[j] = t
            r = r * width + s[mask & (s[i] + s[j])]
            count -= 1

        self.i = i
        self.j = j

        return r
